/* BdfServer_JsonProducers - Copyright (c) 2012-2023 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.jsonproducers.exportation;

import fr.exemole.bdfserver.api.BdfServer;
import fr.exemole.bdfserver.api.instruction.BdfCommandResult;
import fr.exemole.bdfserver.api.instruction.BdfInstructionConstants;
import fr.exemole.bdfserver.api.instruction.OutputParameters;
import fr.exemole.bdfserver.api.interaction.domains.ExportationDomain;
import fr.exemole.bdfserver.jsonproducers.ResultJsonProducer;
import fr.exemole.bdfserver.tools.balayage.engine.BalayageLog;
import fr.exemole.bdfserver.tools.instruction.BdfCommandResultBuilder;
import fr.exemole.bdfserver.tools.instruction.BdfErrors;
import fr.exemole.bdfserver.tools.instruction.BdfInstructionUtils;
import fr.exemole.bdfserver.tools.instruction.BdfOutputUtils;
import java.text.ParseException;
import java.util.LinkedHashSet;
import java.util.Set;
import net.fichotheque.exportation.balayage.BalayageContentDescription;
import net.fichotheque.exportation.balayage.BalayageDescription;
import net.fichotheque.exportation.table.TableExportContentDescription;
import net.fichotheque.exportation.table.TableExportDescription;
import net.fichotheque.exportation.transformation.TransformationKey;
import net.fichotheque.tools.exportation.scrutari.ScrutariExportUtils;
import net.mapeadores.util.json.JsonProducer;
import net.mapeadores.util.json.JsonProperty;
import net.mapeadores.util.json.JsonUtils;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.MessageLocalisation;
import net.mapeadores.util.logging.ErrorMessageException;
import net.mapeadores.util.logging.MessageByLineLog;
import net.mapeadores.util.request.RequestMap;
import net.mapeadores.util.text.StringUtils;


/**
 *
 * @author Vincent Calame
 */
public final class ExportationJsonProducerFactory {

    private ExportationJsonProducerFactory() {
    }

    public static JsonProducer getJsonProducer(OutputParameters parameters) throws ErrorMessageException {
        String json = parameters.getOutput();
        BdfCommandResult bdfCommandResult = parameters.getBdfCommandResult();
        BdfServer bdfServer = parameters.getBdfServer();
        RequestMap requestMap = parameters.getRequestMap();
        if (bdfCommandResult == null) {
            switch (json) {
                case ExportationDomain.SCRUTARIEXPORT_JSON:
                    MessageByLineLog messageByLineLog = ScrutariExportUtils.checkFieldGeneration(BdfOutputUtils.getScrutariExportDef(parameters), bdfServer.getTableExportContext());
                    if (!messageByLineLog.getLogGroupList().isEmpty()) {
                        bdfCommandResult = BdfCommandResultBuilder.init()
                                .putResultObject(BdfInstructionConstants.MESSAGEBYLINELOG_OBJ, messageByLineLog)
                                .toBdfCommandResult();
                    }
            }
        }
        ResultJsonProducer jsonProducer = new ResultJsonProducer(bdfCommandResult, parameters.getMessageLocalisation());
        JsonProperty jsonProperty = getJsonProperty(parameters, json, bdfCommandResult);
        if (jsonProperty == null) {
            return null;
        } else {
            jsonProducer.add(jsonProperty);
            return jsonProducer;
        }
    }

    private static JsonProperty getJsonProperty(OutputParameters parameters, String name, BdfCommandResult bdfCommandResult) throws ErrorMessageException {
        BdfServer bdfServer = parameters.getBdfServer();
        RequestMap requestMap = parameters.getRequestMap();
        Lang lang = parameters.getWorkingLang();
        MessageLocalisation messageLocalisation = parameters.getMessageLocalisation();
        switch (name) {
            case ExportationDomain.TABLEEXPORT_JSON: {
                return new TableExportJsonProperty(BdfOutputUtils.getTableExportDescription(parameters), lang, messageLocalisation);
            }
            case ExportationDomain.TABLEEXPORT_ARRAY_JSON: {
                bdfServer.getTableExportManager().update();
                return new TableExportArrayJsonProperty(bdfServer, lang, messageLocalisation);
            }
            case ExportationDomain.TABLEEXPORT_CONTENT_JSON: {
                return new TableExportContentJsonProperty(bdfServer, getTableExportContentDescription(bdfServer, requestMap), messageLocalisation);
            }
            case ExportationDomain.SCRUTARIEXPORT_JSON: {
                return new ScrutariExportJsonProperty(BdfOutputUtils.getScrutariExportDef(parameters), lang);
            }
            case ExportationDomain.SCRUTARIEXPORT_ARRAY_JSON: {
                return new ScrutariExportArrayJsonProperty(bdfServer, lang);
            }
            case ExportationDomain.SCRUTARIEXPORT_PATHS_JSON: {
                return new ScrutariExportPathsJsonProperty(BdfOutputUtils.getScrutariExportDef(parameters), BdfOutputUtils.getPathConfiguration(parameters));
            }
            case ExportationDomain.TRANSFORMATION_DESCRIPTION_JSON: {
                return new TransformationDescriptionJsonProperty(BdfOutputUtils.getTransformationDescription(parameters), lang, messageLocalisation);
            }
            case ExportationDomain.TRANSFORMATION_DESCRIPTIONS_JSON: {
                bdfServer.getTransformationManager().update();
                return new TransformationDescriptionsJsonProperty(bdfServer.getTransformationManager(), getTransformationKeySet(bdfServer, requestMap), lang, messageLocalisation);
            }
            case ExportationDomain.TRANSFORMATION_TEMPLATEDESCRIPTION_JSON: {
                return new TemplateDescriptionJsonProperty(BdfOutputUtils.getTemplateDescription(parameters), lang, messageLocalisation);
            }
            case ExportationDomain.TRANSFORMATION_TEMPLATECONTENT_JSON: {
                return new TemplateContentJsonProperty(bdfServer, BdfOutputUtils.getTemplateContentDescription(parameters), messageLocalisation);
            }
            case ExportationDomain.SQLEXPORT_JSON: {
                return new SqlExportJsonProperty(BdfOutputUtils.getSqlExportDef(parameters), lang);
            }
            case ExportationDomain.SQLEXPORT_ARRAY_JSON: {
                return new SqlExportArrayJsonProperty(bdfServer, lang);
            }
            case ExportationDomain.SQLEXPORT_PATHS_JSON: {
                return new SqlExportPathsJsonProperty(BdfOutputUtils.getSqlExportDef(parameters), BdfOutputUtils.getPathConfiguration(parameters));
            }
            case ExportationDomain.ACCESS_JSON: {
                return new AccessJsonProperty(BdfOutputUtils.getAccessDef(parameters), lang);
            }
            case ExportationDomain.ACCESS_ARRAY_JSON: {
                return new AccessArrayJsonProperty(bdfServer, lang);
            }
            case ExportationDomain.BALAYAGE_JSON: {
                return new BalayageJsonProperty(BdfOutputUtils.getBalayageDescription(parameters), lang, messageLocalisation);
            }
            case ExportationDomain.BALAYAGE_ARRAY_JSON: {
                bdfServer.getBalayageManager().update();
                return new BalayageArrayJsonProperty(bdfServer, lang, messageLocalisation);
            }
            case ExportationDomain.BALAYAGE_CONTENT_JSON: {
                return new BalayageContentJsonProperty(bdfServer, getBalayageContentDescription(bdfServer, requestMap), messageLocalisation);
            }
            case ExportationDomain.BALAYAGE_LOG_JSON: {
                BalayageLog balayageLog = (BalayageLog) parameters.getResultObject(BdfInstructionConstants.BALAYAGELOG_OBJ);
                if (balayageLog != null) {
                    return balayageLog;
                } else {
                    return JsonUtils.EMPTY_JSONPROPERTY;
                }
            }
            default:
                return null;
        }
    }

    private static BalayageContentDescription getBalayageContentDescription(BdfServer bdfServer, RequestMap requestMap) throws ErrorMessageException {
        BalayageDescription balayageDescription = BdfInstructionUtils.getMandatoryBalayageDescription(bdfServer.getBalayageManager(), requestMap);
        return BdfInstructionUtils.getMandatoryBalayageContentDescription(balayageDescription, requestMap);
    }

    private static TableExportContentDescription getTableExportContentDescription(BdfServer bdfServer, RequestMap requestMap) throws ErrorMessageException {
        TableExportDescription tableExportDescription = BdfInstructionUtils.getMandatoryTableExportDescription(bdfServer.getTableExportManager(), requestMap);
        return BdfInstructionUtils.getMandatoryTableExportContentDescription(tableExportDescription, requestMap);
    }

    private static Set<TransformationKey> getTransformationKeySet(BdfServer bdfServer, RequestMap requestMap) throws ErrorMessageException {
        String[] keysArray = BdfInstructionUtils.getMandatoryParameterValues(requestMap, "keys");
        Set<TransformationKey> transformationKeySet = new LinkedHashSet<TransformationKey>();
        for (String keys : keysArray) {
            String[] tokens = StringUtils.getTechnicalTokens(keys, true);
            for (String token : tokens) {
                try {
                    transformationKeySet.add(TransformationKey.parse(token));
                } catch (ParseException pe) {
                    throw BdfErrors.unknownParameterValue("keys", token);
                }
            }
        }
        return transformationKeySet;
    }

}
